import duckdb
from rich import print


class DataIngestor:
    _s3_data_path: str
    _schema_name: str
    _database_name: str = 'postgresql_db'
    _version_table_name: str = 'version'
    _version_file_name: str = 'version.txt'
    _conn: duckdb.DuckDBPyConnection | None = None

    def __init__(
            self,
            s3_data_path: str,
            schema_name: str,
    ):
        self._s3_data_path = s3_data_path.rstrip('/') + '/'
        self._schema_name = schema_name

    @property
    def _current_conn(self) -> duckdb.DuckDBPyConnection:
        if not self._conn:
            raise RuntimeError('Connection is not established')
        return self._conn

    def connect(self, postgresql_uri: str) -> None:
        if self._conn:
            self._conn.close()
        self._conn = duckdb.connect()
        postgresql_uri = postgresql_uri.replace('postgres://', 'postgresql://')
        self._conn.execute(f"ATTACH '{postgresql_uri}' AS {self._database_name} (TYPE POSTGRES);")

    def load_data_to_database(self, table_name: str):
        s3_path: str = f'{self._s3_data_path}{table_name}.parquet'
        full_table_name: str = self._full_database_table_name(table_name)
        print(
            f'  :arrow_right:  Writing [bold]{s3_path}[/bold] data to [bold]{full_table_name}[/bold] table...',
        )
        # Recreating database table with data
        self._current_conn.execute(f"""
            DROP TABLE IF EXISTS {full_table_name} CASCADE;
            CREATE TABLE {full_table_name} AS FROM '{s3_path}';
        """)

    def create_version_table(self, data_version: str):
        print(f'  :arrow_right:  Updating stored date version to [bold]{data_version}[/bold]...')
        full_table_name: str = self._full_database_table_name(self._version_table_name)
        self._current_conn.execute(f"""
            DROP TABLE IF EXISTS {full_table_name};
            CREATE TABLE {full_table_name} AS 
            SELECT 
                '{data_version}' as data_version,
                now() as updated_at;
        """)

    def get_version(self):
        full_table_name = self._full_database_table_name(self._version_table_name)
        query = f"""
            SELECT data_version
            FROM {full_table_name}
        """
        return self._current_conn.execute(query).fetchone()[0]

    def get_remote_version(self) -> str:
        print(f':cloud:  Getting the latest version of data from remote file storage...')
        s3_path: str = self._s3_data_path + self._version_file_name
        results = self._current_conn.execute(f"SELECT content FROM read_text('{s3_path}')").fetchone()
        if not results:
            raise RuntimeError(f'Unable to get remote version from {s3_path}')
        return str(results[0].strip())

    def create_schema(self):
        print(f':gear:  Creating schema [bold]{self._schema_name}[/bold] for data...')
        self._current_conn.execute(f'CREATE SCHEMA IF NOT EXISTS {self._database_name}.{self._schema_name};')

    def _full_database_table_name(self, table_name: str) -> str:
        return f'"{self._database_name}"."{self._schema_name}"."{table_name}"'

    def __del__(self):
        if self._conn:
            self._conn.close()



from typing import Annotated, List
import typer
from rich import print
from src.data_ingestor import DataIngestor

s3_data_path: str = 'https://inzhenerka-public.s3.eu-west-1.amazonaws.com/scooters_data_generator/'
tables: List[str] = ['trips', 'users', 'events']
default_database_uri: str = 'postgresql://postgres:postgres@localhost:5432/postgres'
default_schema_name: str = 'scooters_raw'

DatabaseURI = Annotated[str, typer.Argument(
    help='Database URI starting with postgresql://',
)]
SchemaName = Annotated[str, typer.Option(
    help='Name of schema in database',
)]

app = typer.Typer(
    name='Scooters Data Uploader',
    help='Upload data to database from remote repository and compare versions',
)


@app.command()
def version(
        database_uri: DatabaseURI = default_database_uri,
        schema_name: SchemaName = default_schema_name
):
    """Compare version of data in database with the latest version available remotely."""
    data_ingestor = DataIngestor(s3_data_path=s3_data_path, schema_name=schema_name)
    data_ingestor.connect(database_uri)
    remote_data_version = data_ingestor.get_remote_version()
    current_data_version = data_ingestor.get_version()
    printc(f"Latest data version available: {remote_data_version}")
    printc(f"Current data version in database: {current_data_version}")
    if not current_data_version:
        printc(":x:  Database is empty. Run 'upload' command to upload data", 'red')
    elif current_data_version != remote_data_version:
        printc(":rotating_light:  Database is outdated. Run 'upload' command to upload data", 'yellow')
    else:
        printc(":white_check_mark:  Database is up to date, nothing to do", 'green')


@app.command()
def upload(
        database_uri: DatabaseURI = default_database_uri,
        schema_name: SchemaName = default_schema_name
):
    """Upload data to database from remote repository."""
    data_ingestor = DataIngestor(s3_data_path=s3_data_path, schema_name=schema_name)
    data_ingestor.connect(database_uri)
    data_version: str = data_ingestor.get_remote_version()
    data_ingestor.create_schema()
    for table_name in tables:
        data_ingestor.load_data_to_database(table_name=table_name)
    data_ingestor.create_version_table(data_version=data_version)
    printc(
        f':white_check_mark:  Data successfully uploaded to schema [bold]{schema_name}[/bold]', 'green'
    )


def printc(msg: str, color: str = 'white'):
    print(f'[{color}]{msg}[/{color}]')


if __name__ == "__main__":
    app()
